/* -*- c-basic-offset: 8 -*-
   rdesktop: A Remote Desktop Protocol client.
   Disk Redirection
   Copyright (C) Jeroen Meijer <jeroen@oldambt7.com> 2003-2008
   Copyright 2003-2011 Peter Astrand <astrand@cendio.se> for Cendio AB
   Copyright 2017 Henrik Andersson <hean01@cendio.se> for Cendio AB
   Copyright 2017 Karl Mikaelsson <derfian@cendio.se> for Cendio AB
   Copyright 2017 Alexander Zakharov <uglym8@gmail.com>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
															
*/

#include "disk.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>		/* open, close */
#include <dirent.h>		/* opendir, closedir, readdir */
#ifdef ENABLE_AMIGA
#include <proto/dos.h>
#else
#include <fnmatch.h>
#endif
#include <errno.h>		/* errno */
#include <stdio.h>

#include <time.h>		/* ctime */

#ifdef ENABLE_AMIGA
#define DIRFD(a) amiga_get_dirfd(a)
#else
#if (defined(HAVE_DIRFD) || (HAVE_DECL_DIRFD == 1))
#define DIRFD(a) (dirfd(a))
#else
#define DIRFD(a) ((a)->DIR_FD_MEMBER_NAME)
#endif
#endif

/* TODO: Fix mntent-handling for solaris
 * #include <sys/mntent.h> */
#if (defined(HAVE_MNTENT_H) && defined(HAVE_SETMNTENT))
#include <mntent.h>
#define MNTENT_PATH "/etc/mtab"
#define USE_SETMNTENT
#endif

#ifdef HAVE_SYS_VFS_H
#include <sys/vfs.h>
#endif

#ifdef HAVE_SYS_STATVFS_H
#include <sys/statvfs.h>
#endif

#ifdef HAVE_SYS_STATFS_H
#include <sys/statfs.h>
#endif

#ifdef HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif

#ifdef HAVE_SYS_MOUNT_H
#include <sys/mount.h>
#endif

#include "rdesktop.h"

extern void rdp_out_unistr(STREAM s, char *string, int len);

#ifdef STAT_STATFS3_OSF1
#define STATFS_FN(path, buf) (statfs(path,buf,sizeof(buf)))
#define STATFS_T statfs
#define USE_STATFS
#endif

#ifdef STAT_STATVFS
#define STATFS_FN(path, buf) (statvfs(path,buf))
#define STATFS_T statvfs
#define USE_STATVFS
#endif

#ifdef STAT_STATVFS64
#define STATFS_FN(path, buf) (statvfs64(path,buf))
#define STATFS_T statvfs64
#define USE_STATVFS
#endif

#if (defined(STAT_STATFS2_FS_DATA) || defined(STAT_STATFS2_BSIZE) || defined(STAT_STATFS2_FSIZE))
#define STATFS_FN(path, buf) (statfs(path,buf))
#define STATFS_T statfs
#define USE_STATFS
#endif

#ifdef STAT_STATFS4
#define STATFS_FN(path, buf) (statfs(path,buf,sizeof(buf),0))
#define STATFS_T statfs
#define USE_STATFS
#endif

#if ((defined(USE_STATFS) && defined(HAVE_STRUCT_STATFS_F_NAMEMAX)) || (defined(USE_STATVFS) && defined(HAVE_STRUCT_STATVFS_F_NAMEMAX)))
#define F_NAMELEN(buf) ((buf).f_namemax)
#endif

#if ((defined(USE_STATFS) && defined(HAVE_STRUCT_STATFS_F_NAMELEN)) || (defined(USE_STATVFS) && defined(HAVE_STRUCT_STATVFS_F_NAMELEN)))
#define F_NAMELEN(buf) ((buf).f_namelen)
#endif

#ifndef F_NAMELEN
#define F_NAMELEN(buf) (255)
#endif

/* Dummy statfs fallback */
#ifndef STATFS_T
struct dummy_statfs_t
{
	long f_bfree;
	long f_bsize;
	long f_bavail;
	long f_blocks;
	int f_namelen;
	int f_namemax;

};

static int
dummy_statfs(struct dummy_statfs_t *buf)
{
	buf->f_blocks = 262144;
	buf->f_bfree = 131072;
	buf->f_bavail = 131072;
	buf->f_bsize = 512;
	buf->f_namelen = 255;
	buf->f_namemax = 255;


	return 0;
}

#define STATFS_T dummy_statfs_t
#define STATFS_FN(path,buf) (dummy_statfs(buf))
#endif

extern RDPDR_DEVICE g_rdpdr_device[];

#ifdef ENABLE_AMIGA
FILEINFO g_fileinfo[MAX_OPEN_FILES*2];
#else
FILEINFO g_fileinfo[MAX_OPEN_FILES];
#endif
RD_BOOL g_notify_stamp = False;

struct DateStamp convertTimeT(time_t time_epoch)
{
	struct DateStamp date;
	time_epoch -=  252460800;
	
	date.ds_Days = time_epoch%86400;
	date.ds_Minute = time_epoch%60;
	date.ds_Tick = 0;
	return date;
}


time_t convertDateStamp(struct DateStamp *date)
{
	return date->ds_Days*86400+date->ds_Minute*60+date->ds_Tick/TICKS_PER_SECOND+252460800;
}


typedef struct
{
	char name[PATH_MAX];
	char label[PATH_MAX];
	unsigned long serial;
	char type[PATH_MAX];
} FsInfoType;

static RD_NTSTATUS NotifyInfo(RD_NTHANDLE handle, uint32 info_class, NOTIFY * p);

static time_t
get_create_time(struct stat *filestat)
{
	time_t ret, ret1;

	ret = MIN(filestat->st_ctime, filestat->st_mtime);
	ret1 = MIN(ret, filestat->st_atime);

	if (ret1 != (time_t) 0)
		return ret1;

	return ret;
}

/* Convert seconds since 1970 to a filetime */
static void
seconds_since_1970_to_filetime(time_t seconds, uint32 * high, uint32 * low)
{
	unsigned long long ticks;

	ticks = (seconds + 11644473600LL) * 10000000;
	*low = (uint32) ticks;
	*high = (uint32) (ticks >> 32);
}

/* Convert seconds since 1970 back to filetime */
static time_t
convert_1970_to_filetime(uint32 high, uint32 low)
{
	unsigned long long ticks;
	time_t val;

	ticks = low + (((unsigned long long) high) << 32);
	ticks /= 10000000;
	ticks -= 11644473600LL;

	val = (time_t) ticks;
	return (val);

}

/* A wrapper for ftruncate which supports growing files, even if the
   native ftruncate doesn't. This is needed on Linux FAT filesystems,
   for example. */
static int
ftruncate_growable(int fd, off_t length)
{
	int ret;
	off_t pos;
	static const char zero = 0;

	/* Try the simple method first */
	if ((ret = ftruncate(fd, length)) != -1)
	{
		return ret;
	}

	/*
	 * Some kind of error. Perhaps we were trying to grow. Retry
	 * in a safe way.
	 */

	/* Get current position */
	if ((pos = lseek(fd, 0, SEEK_CUR)) == -1)
	{
		logger(Disk, RDPDebug, "ftruncate_growable(), lseek() failed: %s", strerror(errno));
		return -1;
	}

	/* Seek to new size */
	if (lseek(fd, length, SEEK_SET) == -1)
	{
		logger(Disk, RDPDebug, "ftruncate_growable(), lseek() failed: %s", strerror(errno));
		return -1;
	}

	/* Write a zero */
	if (write(fd, &zero, 1) == -1)
	{
		logger(Disk, RDPDebug, "ftruncate_growable(), write() failed: %s", strerror(errno));
		return -1;
	}

	/* Truncate. This shouldn't fail. */
	if (ftruncate(fd, length) == -1)
	{
		logger(Disk, RDPDebug, "ftruncate_growable(), ftruncate() failed: %s",
		       strerror(errno));
		return -1;
	}

	/* Restore position */
	if (lseek(fd, pos, SEEK_SET) == -1)
	{
		logger(Disk, RDPDebug, "ftruncate_growable(), lseek() failed: %s", strerror(errno));
		return -1;
	}

	return 0;
}

/* Just like open(2), but if a open with O_EXCL fails, retry with
   GUARDED semantics. This might be necessary because some filesystems
   (such as NFS filesystems mounted from a unfsd server) doesn't
   support O_EXCL. GUARDED semantics are subject to race conditions,
   but we can live with that.
*/
static int
open_weak_exclusive(const char *pathname, int flags, mode_t mode)
{
	int ret;
	struct stat filestat;

	ret = open(pathname, flags, mode);
	if (ret != -1 || !(flags & O_EXCL))
	{
		/* Success, or not using O_EXCL */
		return ret;
	}

	/* An error occurred, and we are using O_EXCL. In case the FS
	   doesn't support O_EXCL, some kind of error will be
	   returned. Unfortunately, we don't know which one. Linux
	   2.6.8 seems to return 524, but I cannot find a documented
	   #define for this case. So, we'll return only on errors that
	   we know aren't related to O_EXCL. */
	switch (errno)
	{
		case EACCES:
		case EEXIST:
		case EINTR:
		case EISDIR:
		case ELOOP:
		case ENAMETOOLONG:
		case ENOENT:
		case ENOTDIR:
			return ret;
	}

	/* Retry with GUARDED semantics */
	if (stat(pathname, &filestat) != -1)
	{
		/* File exists */
		errno = EEXIST;
		return -1;
	}
	else
	{
		return open(pathname, flags & ~O_EXCL, mode);
	}
}

/* Enumeration of devices from rdesktop.c        */
/* returns number of units found and initialized. */
/* optarg looks like ':h=/mnt/floppy,b=/mnt/usbdevice1' */
/* when it arrives to this function.             */
int
disk_enum_devices(uint32 * id, char *optarg)
{
	char *pos = optarg;
	char *pos2;
	int count = 0;
	DISK_DEVICE *pdisk_data;

#ifndef ENABLE_AMIGA
	//KPrintF("optarg: %s\n", optarg);
	// skip the first colon 
	optarg++;
#endif
	while ((pos = next_arg(optarg, ',')) && *id < RDPDR_MAX_DEVICES)
	{
		pos2 = next_arg(optarg, '=');

		pdisk_data = (DISK_DEVICE *) xmalloc(sizeof(DISK_DEVICE));
		memset(pdisk_data, 0, sizeof(DISK_DEVICE));
		strncpy(pdisk_data->name, optarg, sizeof(pdisk_data->name) - 1);
		strncpy(g_rdpdr_device[*id].name, optarg, sizeof(g_rdpdr_device[*id].name) - 1);

		g_rdpdr_device[*id].local_path = (char *) xmalloc(strlen(pos2) + 1);
		strcpy(g_rdpdr_device[*id].local_path, pos2);
		g_rdpdr_device[*id].device_type = DEVICE_TYPE_DISK;
		g_rdpdr_device[*id].pdevice_data = (void *) pdisk_data;

		count++;
		(*id)++;

		optarg = pos;
	}
	return count;
}

#ifdef ENABLE_AMIGA
int amiga_get_dirfd(BPTR dirp)
{
  int i;

  for (i = MAX_OPEN_FILES ; i < MAX_OPEN_FILES*2 ; i++)
  {
    if (NULL == g_fileinfo[i].pdir)
    {
       return i - MAX_OPEN_FILES;
    }
  }

  return MAX_OPEN_FILES;
}
#endif

/* Opens or creates a file or directory */
static RD_NTSTATUS
disk_create(uint32 device_id, uint32 accessmask, uint32 sharemode, uint32 create_disposition,
	    uint32 flags_and_attributes, char *filename, RD_NTHANDLE * phandle)
{
	RD_NTHANDLE handle;
	BPTR dirp = NULL;
	int flags, mode;
	char path[PATH_MAX];
	struct stat filestat;

	if (flags_and_attributes == 0x20 && filename && filename[0] == '/' && filename[1] == 0)
	{
		return RD_STATUS_INVALID_PARAMETER;
	}

	if (!filename)
	{
		//KPrintF("%s: filename is NULL !\n",__FUNCTION__);
		return RD_STATUS_NO_SUCH_FILE;
	}

	if (strchr(filename, ':')) // filename:Zone.Identifier, directory:SummaryInformation:$DATA, etc.
	{
		return RD_STATUS_NO_SUCH_FILE;
	}

	if (filename && filename[0] == '/')
	{
		filename++;
	}

	handle = 0;
	flags = 0;
	mode = S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;

	if (filename && *filename && filename[strlen(filename) - 1] == '/')
		filename[strlen(filename) - 1] = 0;
#ifdef ENABLE_AMIGA
	strcpy(path, g_rdpdr_device[device_id].local_path);
	AddPart(path,  filename ? filename : "", PATH_MAX);
#else
	sprintf(path, "%s%s", g_rdpdr_device[device_id].local_path,  filename ? filename : ""));
#endif

	switch (create_disposition)
	{
		case CREATE_ALWAYS:
			//KPrintF("CREATE_ALWAYS\n");
			/* Delete existing file/link. */
			DeleteFile(path);
			flags |= O_CREAT;
			break;

		case CREATE_NEW:
			//KPrintF("CREATE_NEW\n");
			/* If the file already exists, then fail. */
			flags |= O_CREAT | O_EXCL;
			break;

		case OPEN_ALWAYS:

			/* Create if not already exists. */
			flags |= O_CREAT;
			break;

		case OPEN_EXISTING:

			/* Default behaviour */
			break;

		case TRUNCATE_EXISTING:

			/* If the file does not exist, then fail. */
			flags |= O_TRUNC;
			break;
	}

	//KPrintF("Open: %s   \"%s\"  flags: %X, accessmask: %X sharemode: %X create disp: %X\n", filename, path, flags_and_attributes, accessmask, sharemode, create_disposition); 

	/* Get information about file and set that flag ourselves */
	if ((stat(path, &filestat) == 0) && (S_ISDIR(filestat.st_mode)))
	{
		if (flags_and_attributes & FILE_NON_DIRECTORY_FILE)
			return RD_STATUS_FILE_IS_A_DIRECTORY;
		else
			flags_and_attributes |= FILE_DIRECTORY_FILE;
	}

	if (flags_and_attributes & FILE_DIRECTORY_FILE)
	{
		if (flags & O_CREAT)
		{
			//KPrintF("Creating Dir\n");
			dirp = CreateDir(path);
			
			UnLock(dirp);
		}
	
		dirp = Lock(path, SHARED_LOCK);
			
		if (!dirp)
		{
			LONG amiga_error  = IoErr();
			//KPrintF("Amiga error: %ld\n", amiga_error);
			switch (amiga_error)
			{
				case ERROR_READ_PROTECTED:
					return RD_STATUS_ACCESS_DENIED;
				case ERROR_DIR_NOT_FOUND:
				case ERROR_OBJECT_NOT_FOUND:
					return RD_STATUS_NO_SUCH_FILE;
				default:
					return RD_STATUS_NO_SUCH_FILE;
			}
		}
		handle = DIRFD(dirp);
	}
	else
	{
		if (accessmask & GENERIC_ALL
		    || (accessmask & GENERIC_READ && accessmask & GENERIC_WRITE))
		{
			flags |= O_RDWR;
		}
		else if ((accessmask & GENERIC_WRITE) && !(accessmask & GENERIC_READ))
		{
			flags |= O_WRONLY;
		}
		else
		{
			flags |= O_RDONLY;
		}

		handle = open_weak_exclusive(path, flags, mode);
		if (handle == -1)
		{
			switch (errno)
			{
				case EISDIR:
					return RD_STATUS_FILE_IS_A_DIRECTORY;
				case EACCES:
					return RD_STATUS_ACCESS_DENIED;

				case ENOENT:
					return RD_STATUS_NO_SUCH_FILE;
				case EEXIST:
					return RD_STATUS_OBJECT_NAME_COLLISION;
				default:
					logger(Disk, RDPDebug, "disk_create(), open() failed: %s",
					       strerror(errno));
					return RD_STATUS_NO_SUCH_FILE;
			}
		}
	}

	if (handle >= MAX_OPEN_FILES)
	{
		//KPrintF("Maximum number of open files (%d) reached. Increase MAX_OPEN_FILES!\n", handle);
		exit(1);
	}

	if (dirp)
	{
		handle += MAX_OPEN_FILES;
		g_fileinfo[handle].pdir = dirp;
	}
	else
		g_fileinfo[handle].pdir = NULL;
		
	g_fileinfo[handle].FIB = AllocDosObject(DOS_FIB, NULL);
	
	g_fileinfo[handle].device_id = device_id;
	g_fileinfo[handle].flags_and_attributes = flags_and_attributes;
	g_fileinfo[handle].accessmask = accessmask;
	strncpy(g_fileinfo[handle].path, path, PATH_MAX - 1);
	g_fileinfo[handle].delete_on_close = False;

	if (accessmask & GENERIC_ALL || accessmask & GENERIC_WRITE)
		g_notify_stamp = True;

	*phandle = handle;
    //KPrintF("Disk create ok\n");
	return RD_STATUS_SUCCESS;
}

static RD_NTSTATUS
disk_close(RD_NTHANDLE handle)
{
	struct fileinfo *pfinfo;
	//KPrintF("disk_close\n");
	pfinfo = &(g_fileinfo[handle]);

	if (pfinfo->accessmask & GENERIC_ALL || pfinfo->accessmask & GENERIC_WRITE)
		g_notify_stamp = True;

	rdpdr_abort_io(handle, 0, RD_STATUS_CANCELLED);

	if (pfinfo->pdir)
	{
		UnLock(pfinfo->pdir);
		pfinfo->pdir = NULL;
	}
	else
	{
		if (close(handle) < 0)
		{
			perror("close");
			return RD_STATUS_INVALID_HANDLE;
		}
	}
	
	if (g_fileinfo[handle].FIB)
	{
			FreeDosObject(DOS_FIB, g_fileinfo[handle].FIB);
			g_fileinfo[handle].FIB =  NULL;
	}
	
	if (pfinfo->delete_on_close)
	{
		if (DeleteFile(pfinfo->path) == FALSE)
		{
			//KPrintF("Error deleting: %s\n", pfinfo->path);
			return RD_STATUS_ACCESS_DENIED;
		}

		pfinfo->delete_on_close = False;
	}

	return RD_STATUS_SUCCESS;
}

static RD_NTSTATUS
disk_read(RD_NTHANDLE handle, uint8 * data, uint32 length, uint64 offset, uint32 * result)
{
	int n;
#if 0
	/* browsing dir ????        */
	/* each request is 24 bytes */
	if (g_fileinfo[handle].flags_and_attributes & FILE_DIRECTORY_FILE)
	{
		*result = 0;
		return STATUS_SUCCESS;
	}
#endif
#ifdef ENABLE_AMIGA
	//KPrintF("Disk read\n");
	if (g_fileinfo[handle].flags_and_attributes & FILE_DIRECTORY_FILE)
	{
		*result = 0;
		return RD_STATUS_NOT_IMPLEMENTED;
	}
#endif

	lseek(handle, offset, SEEK_SET);

	n = read(handle, data, length);

	if (n < 0)
	{
		*result = 0;
		switch (errno)
		{
			case EISDIR:
				/* Implement 24 Byte directory read ??
				   with STATUS_NOT_IMPLEMENTED server doesn't read again */
				/* return STATUS_FILE_IS_A_DIRECTORY; */
				return RD_STATUS_NOT_IMPLEMENTED;
			default:
				logger(Disk, RDPDebug, "disk_read(), read failed: %s",
				       strerror(errno));
				return RD_STATUS_INVALID_PARAMETER;
		}
	}

	*result = n;

	return RD_STATUS_SUCCESS;
}

static RD_NTSTATUS
disk_write(RD_NTHANDLE handle, uint8 * data, uint32 length, uint64 offset, uint32 * result)
{
	int n;
	//KPrintF("Disk write\n");
	lseek(handle, offset, SEEK_SET);

	n = write(handle, data, length);

	if (n < 0)
	{
		logger(Disk, RDPDebug, "disk_write(), write() failed: %s", strerror(errno));
		*result = 0;
		switch (errno)
		{
			case ENOSPC:
				return RD_STATUS_DISK_FULL;
			default:
				return RD_STATUS_ACCESS_DENIED;
		}
	}

	*result = n;

	return RD_STATUS_SUCCESS;
}

/* Btw, all used Flie* structures are described in [MS-FSCC] */
RD_NTSTATUS
disk_query_information(RD_NTHANDLE handle, uint32 info_class, STREAM out)
{
	uint32 file_attributes, ft_high, ft_low;
	char *path, *filename;
	int aa=0;
	struct FileInfoBlock *fib;
	time_t date;
	BPTR a_handle;
	path = g_fileinfo[handle].path;
	//KPrintF("disk_query_information() File: %s\n", path);
	if ((a_handle = Lock(path, ACCESS_READ)))
	{
		fib = g_fileinfo[handle].FIB;
		Examine(a_handle, fib);
	}
	else
	{
		out_uint8(out, 0);
		return RD_STATUS_ACCESS_DENIED;
	}
	date=convertDateStamp(&fib->fib_Date); 

	/* Set file attributes */
	file_attributes = 0;
	if ((fib->fib_DirEntryType == ST_USERDIR) || (fib->fib_DirEntryType == ST_ROOT))
		file_attributes |= FILE_ATTRIBUTE_DIRECTORY;
	

	filename = 	FilePart(path);
	if (filename && filename[0] == '.')
		file_attributes |= FILE_ATTRIBUTE_HIDDEN;

	if (!file_attributes)
		file_attributes |= FILE_ATTRIBUTE_NORMAL;
	
	//KPrintF("a: %ld\n", aa++);
	/* Return requested data */
	switch (info_class)
	{
		case FileBasicInformation:

			seconds_since_1970_to_filetime(date, &ft_high,
						       &ft_low);
			out_uint32_le(out, ft_low);	/* create_access_time */
			out_uint32_le(out, ft_high);

			out_uint32_le(out, ft_low);	/* last_access_time */
			out_uint32_le(out, ft_high);	
			out_uint32_le(out, ft_low);	/* last_write_time */
			out_uint32_le(out, ft_high);

			out_uint32_le(out, ft_low);	/* last_change_time */
			out_uint32_le(out, ft_high);
			
			out_uint32_le(out, file_attributes);			
			break;

		case FileStandardInformation:

			//	printf("FileStandardInformation.: \n");
			out_uint32_le(out, fib->fib_Size);	/* Allocation size */
			out_uint32_le(out, 0);
			out_uint32_le(out, fib->fib_Size);	/* End of file */
			out_uint32_le(out, 0);
			out_uint32_le(out, 1);	/* Number of links */
			out_uint8(out, 0);	/* Delete pending */
			out_uint8(out, (fib->fib_DirEntryType == ST_USERDIR ) ? 1 : 0);	/* Directory */
			break;

		case FileObjectIdInformation:

			out_uint32_le(out, file_attributes);	/* File Attributes */
			out_uint32_le(out, 0);	/* Reparse Tag */
			break;

		default:

			unimpl("IRP Query (File) Information class: 0x%x\n", info_class);
			return RD_STATUS_INVALID_PARAMETER;
	}
	//KPrintF("a: %ld\n", aa++);

	UnLock(a_handle);
	return RD_STATUS_SUCCESS;
}


RD_NTSTATUS
disk_set_information(RD_NTHANDLE handle, uint32 info_class, STREAM in, STREAM out)
{
	uint32 length, file_attributes, ft_high, ft_low, delete_on_close;
	char *newname = NULL, fullpath[PATH_MAX];
	struct fileinfo *pfinfo;
	int mode;
	time_t write_time, change_time, access_time, mod_time;
	struct STATFS_T stat_fs;
	struct FileInfoBlock *fib;
	BPTR a_handle;
    //KPrintF("disk_set_information: %s\n", g_fileinfo[handle].path);

	if ((a_handle = Lock(g_fileinfo[handle].path, ACCESS_READ)))
	{
		pfinfo = &(g_fileinfo[handle]);
		fib = g_fileinfo[handle].FIB;
		g_notify_stamp = True;

		switch (info_class)
		{
			case FileBasicInformation:
				write_time = change_time = access_time = 0;

				in_uint8s(in, 4);	/* length of SetBuffer */
				in_uint8s(in, 24);	/* padding */

				/* CreationTime */
				in_uint32_le(in, ft_low);
				in_uint32_le(in, ft_high);

				/* AccessTime */
				in_uint32_le(in, ft_low);
				in_uint32_le(in, ft_high);
				if (ft_low || ft_high)
					access_time = convert_1970_to_filetime(ft_high, ft_low);

				/* WriteTime */
				in_uint32_le(in, ft_low);
				in_uint32_le(in, ft_high);
				if (ft_low || ft_high)
					write_time = convert_1970_to_filetime(ft_high, ft_low);

				/* ChangeTime */
				in_uint32_le(in, ft_low);
				in_uint32_le(in, ft_high);
				if (ft_low || ft_high)
					change_time = convert_1970_to_filetime(ft_high, ft_low);

				in_uint32_le(in, file_attributes);
				
				if(!Examine(a_handle, fib))
					return RD_STATUS_ACCESS_DENIED;
				
				if (write_time || change_time)
					mod_time = MIN(write_time, change_time);
				else
					mod_time = write_time ? write_time : change_time;

				if (access_time || write_time || change_time)
				{
					struct DateStamp d_stamp = convertTimeT(mod_time);
					if (0==SetFileDate(pfinfo->path, &d_stamp )) 
						return RD_STATUS_ACCESS_DENIED;
				}

				if (!file_attributes)
					break;	/* not valid */

				mode = fib->fib_Protection;

				if (file_attributes & FILE_ATTRIBUTE_READONLY)
					mode &= ~( FIBF_READ );
				else
					mode |= FIBF_WRITE;
				
				SetProtection (pfinfo->path, mode);

				break;

			case FileRenameInformation:
				//KPrintF("Rename\n");
				in_uint8s(in, 4);	/* Handle of root dir? */
				in_uint8s(in, 0x1a);	/* unknown */
				in_uint32_le(in, length);

				if (length && (length / 2) >= PATH_MAX)
					return RD_STATUS_INVALID_PARAMETER;
					
				rdp_in_unistr(in, length, &newname, &length);
				
				if (newname == NULL)
					return RD_STATUS_INVALID_PARAMETER;
				
				convert_to_unix_filename(newname);

	#ifdef ENABLE_AMIGA
				strcpy(fullpath, g_rdpdr_device[pfinfo->device_id].local_path);
				if ('/' == newname[0])
				{
					AddPart(fullpath, &newname[1], PATH_MAX);
				} else {
					AddPart(fullpath, newname, PATH_MAX);
				}
	#else
				
				sprintf(fullpath, "%s%s", g_rdpdr_device[pfinfo->device_id].local_path,
					newname);
	#endif
  			    xfree(newname);             
				if (Rename(pfinfo->path, fullpath) == DOSFALSE)
				{
					//KPrintF("RD_STATUS_ACCESS_DENIED\n");
					return RD_STATUS_ACCESS_DENIED;
				}
				break;

			case FileDispositionInformation:
				/* As far as I understand it, the correct
				   thing to do here is to *schedule* a delete,
				   so it will be deleted when the file is
				   closed. Subsequent
				   FileDispositionInformation requests with
				   DeleteFile set to FALSE should unschedule
				   the delete. See
				   http://www.osronline.com/article.cfm?article=245. */
		
				/* FileDispositionInformation always sets delete_on_close to true.
				   "STREAM in" includes Length(4bytes) , Padding(24bytes) and SetBuffer(zero byte).
				   Length is always set to zero.
				   [MS-RDPEFS] http://msdn.microsoft.com/en-us/library/cc241305%28PROT.10%29.aspx
				   - 2.2.3.3.9 Server Drive Set Information Request
				*/
				in_uint8s(in, 4);	/* length of SetBuffer */
				in_uint8s(in, 24);	/* padding */

				if ((pfinfo->accessmask &                                                                                                    
					 (FILE_DELETE_ON_CLOSE | FILE_COMPLETE_IF_OPLOCKED)))                                                                    
																																			 
				{                                                                                                                            
						/* if file exists in directory , necessary to return RD_STATUS_DIRECTORY_NOT_EMPTY with win2008                      
						   [MS-RDPEFS] http://msdn.microsoft.com/en-us/library/cc241305%28PROT.10%29.aspx                                    
						   - 2.2.3.3.9 Server Drive Set Information Request                                                                  
						   - 2.2.3.4.9 Client Drive Set Information Response                                                                 
						   [MS-FSCC] http://msdn.microsoft.com/en-us/library/cc231987%28PROT.10%29.aspx                                      
						   - 2.4.11 FileDispositionInformation                                                                               
						   [FSBO] http://msdn.microsoft.com/en-us/library/cc246487%28PROT.13%29.aspx                                         
						   - 4.3.2 Set Delete-on-close using FileDispositionInformation Information Class (IRP_MJ_SET_INFORMATION)           
						 */                                                                                                                  
						if (pfinfo->pdir)                                                                                                    
						{                                                                                                                    
								DIR *dp = opendir(pfinfo->path);                                                                             
								struct dirent *dir;                                                                                          
																																			 
								while ((dir = readdir(dp)) != NULL)                                                                          
								{                                                                                                            
										if (strcmp(dir->d_name, ".") != 0                                                                    
											&& strcmp(dir->d_name, "..") != 0)                                                               
										{                                                                                                    
												closedir(dp);                                                                                
												return RD_STATUS_DIRECTORY_NOT_EMPTY;                                                        
										}                                                                                                    
								}                                                                                                            
								closedir(dp);                                                                                                
						}                                                                                                                    
																																			 
						pfinfo->delete_on_close = True;                                                                                      
				}                                                         
				break;

			case FileAllocationInformation:
				/* Fall through to FileEndOfFileInformation,
				   which uses ftrunc. This is like Samba with
				   "strict allocation = false", and means that
				   we won't detect out-of-quota errors, for
				   example. */

			case FileEndOfFileInformation:
				in_uint8s(in, 28);	/* unknown */
				in_uint32_le(in, length);	/* file size */

				/* prevents start of writing if not enough space left on device */
				if (STATFS_FN(pfinfo->path, &stat_fs) == 0)
					if (stat_fs.f_bfree * stat_fs.f_bsize < length)
						return RD_STATUS_DISK_FULL;

				if (ftruncate_growable(handle, length) != 0)
				{
					return RD_STATUS_DISK_FULL;
				}

				break;
			default:

				unimpl("IRP Set File Information class: 0x%x\n", info_class);
				return RD_STATUS_INVALID_PARAMETER;
		}
		UnLock(a_handle);
	}
	else
		return RD_STATUS_ACCESS_DENIED;
	
	
	//KPrintF("disk_set_information() RD_STATUS_SUCCESS\n");
	return RD_STATUS_SUCCESS;
}

RD_NTSTATUS
disk_check_notify(RD_NTHANDLE handle)
{
	struct fileinfo *pfinfo;
	RD_NTSTATUS status = RD_STATUS_PENDING;
	NOTIFY notify;

	//KPrintF("disk_check_notify\n");
	
	pfinfo = &(g_fileinfo[handle]);
	if (!pfinfo->pdir)
		return RD_STATUS_INVALID_DEVICE_REQUEST;

	status = NotifyInfo(handle, pfinfo->info_class, &notify);

	if (status != RD_STATUS_PENDING)
		return status;

	if (memcmp(&pfinfo->notify, &notify, sizeof(NOTIFY)))
	{
		/*printf("disk_check_notify found changed event\n"); */
		memcpy(&pfinfo->notify, &notify, sizeof(NOTIFY));
		status = RD_STATUS_NOTIFY_ENUM_DIR;
	}

	return status;
}

RD_NTSTATUS
disk_create_notify(RD_NTHANDLE handle, uint32 info_class)
{

	struct fileinfo *pfinfo;
	RD_NTSTATUS ret = RD_STATUS_PENDING;

	//KPrintF("start disk_create_notify info_class %X\n", info_class); 

	pfinfo = &(g_fileinfo[handle]);
	pfinfo->info_class = info_class;
#ifdef ENABLE_AMIGA
	g_notify_stamp = False;
	return RD_STATUS_PENDING;
#endif

	ret = NotifyInfo(handle, info_class, &pfinfo->notify);

	if (info_class & 0x1000)
	{			/* ???? */
		if (ret == RD_STATUS_PENDING)
			return RD_STATUS_SUCCESS;
	}

	/* printf("disk_create_notify: num_entries %d\n", pfinfo->notify.num_entries); */


	return ret;

}

static RD_NTSTATUS
NotifyInfo(RD_NTHANDLE handle, uint32 info_class, NOTIFY * p)
{
	struct fileinfo *pfinfo;
	struct stat filestat;
	struct dirent *dp;
	char *fullname;
	DIR *dpr;

	pfinfo = &(g_fileinfo[handle]);
	
	//KPrintF("NotifyInfo %s\n", pfinfo->path);
#ifdef ENABLE_AMIGA
	if (stat(pfinfo->path, &filestat) < 0)
#else
	if (fstat(handle, &filestat) < 0)
#endif
	{
		//KPrintF("NotifyInfo error\n");
		return RD_STATUS_ACCESS_DENIED;
	}
	p->modify_time = filestat.st_mtime;
	p->status_time = filestat.st_ctime;
	p->num_entries = 0;
	p->total_time = 0;


	dpr = opendir(pfinfo->path);
	if (!dpr)
	{
		//KPrintF("NotifyInfo error1\n");
		return RD_STATUS_ACCESS_DENIED;
	}

	while ((dp = readdir(dpr)))
	{
		if (!strcmp(dp->d_name, ".") || !strcmp(dp->d_name, ".."))
			continue;
		p->num_entries++;
		fullname = (char *) xmalloc(strlen(pfinfo->path) + strlen(dp->d_name) + 2);
#ifdef ENABLE_AMIGA
		strcpy(fullname, pfinfo->path);
		AddPart(fullname, dp->d_name, strlen(pfinfo->path) + strlen(dp->d_name) + 2);
#else
		sprintf(fullname, "%s/%s", pfinfo->path, dp->d_name);
#endif

		if (!stat(fullname, &filestat))
		{
			p->total_time += (filestat.st_mtime + filestat.st_ctime);
		}

		xfree(fullname);
	}
	closedir(dpr);

	return RD_STATUS_PENDING;
}

static FsInfoType *
FsVolumeInfo(char *fpath)
{
	static FsInfoType info;
#ifdef USE_SETMNTENT
	FILE *fdfs;
	struct mntent *e;
#endif

	/* initialize */
	memset(&info, 0, sizeof(info));
	strcpy(info.label, "RDESKTOP");
	strcpy(info.type, "RDPFS");

#ifdef USE_SETMNTENT
	fdfs = setmntent(MNTENT_PATH, "r");
	if (!fdfs)
		return &info;

	while ((e = getmntent(fdfs)))
	{
		if (str_startswith(e->mnt_dir, fpath))
		{
			strcpy(info.type, e->mnt_type);
			strcpy(info.name, e->mnt_fsname);
			if (strstr(e->mnt_opts, "vfat") || strstr(e->mnt_opts, "iso9660"))
			{
				int fd = open(e->mnt_fsname, O_RDONLY);
				if (fd >= 0)
				{
					unsigned char buf[512];
					memset(buf, 0, sizeof(buf));
					if (strstr(e->mnt_opts, "vfat"))
						 /*FAT*/
					{
						strcpy(info.type, "vfat");
						read(fd, buf, sizeof(buf));
						info.serial =
							(buf[42] << 24) + (buf[41] << 16) +
							(buf[40] << 8) + buf[39];
						strncpy(info.label, (char *) buf + 43, 10);
						info.label[10] = '\0';
					}
					else if (lseek(fd, 32767, SEEK_SET) >= 0)	/* ISO9660 */
					{
						read(fd, buf, sizeof(buf));
						strncpy(info.label, (char *) buf + 41, 32);
						info.label[32] = '\0';
						/* info.Serial = (buf[128]<<24)+(buf[127]<<16)+(buf[126]<<8)+buf[125]; */
					}
					close(fd);
				}
			}
		}
	}
	endmntent(fdfs);
#else
	/* initialize */
	memset(&info, 0, sizeof(info));
	strcpy(info.label, "RDESKTOP");
	strcpy(info.type, "RDPFS");

#endif
	return &info;
}


RD_NTSTATUS
disk_query_volume_information(RD_NTHANDLE handle, uint32 info_class, STREAM out)
{
	struct STATFS_T stat_fs;
	struct fileinfo *pfinfo;
	FsInfoType *fsinfo;
    STREAM stmp;     
	
    //KPrintF("disk_query_volume_information \n");    
	
	pfinfo = &(g_fileinfo[handle]);

	if (STATFS_FN(pfinfo->path, &stat_fs) != 0)
	{
		perror("statfs");
		//KPrintF("statfs\n");
		return RD_STATUS_ACCESS_DENIED;
	}

	fsinfo = FsVolumeInfo(pfinfo->path);

	switch (info_class)
	{
		case FileFsVolumeInformation:
           	stmp = s_alloc(PATH_MAX * 4);
			out_utf16s(stmp, fsinfo->label);
			s_mark_end(stmp);                                                                                                         

			out_uint32_le(out, 0);	/* volume creation time low */
			out_uint32_le(out, 0);	/* volume creation time high */
			out_uint32_le(out, fsinfo->serial);	/* serial */
     	    out_uint32_le(out, s_length(stmp));    /* length of string */                                                               
            out_uint8(out, 0);      /* support objects? */                                                                               
            out_stream(out, stmp);  /* fsinfo->label string*/           		
			s_free(stmp);			 
    	break;

		case FileFsSizeInformation:

			out_uint32_le(out, stat_fs.f_blocks);	/* Total allocation units low */
			out_uint32_le(out, 0);	/* Total allocation high units */
			out_uint32_le(out, stat_fs.f_bfree);	/* Available allocation units */
			out_uint32_le(out, 0);	/* Available allowcation units */
			out_uint32_le(out, stat_fs.f_bsize / 0x200);	/* Sectors per allocation unit */
			out_uint32_le(out, 0x200);	/* Bytes per sector */
	   break;
       case FileFsFullSizeInformation:                                                                                                      

			out_uint64_le(out, stat_fs.f_blocks);   /* Total allocation units */                                                         
			out_uint64_le(out, stat_fs.f_bavail);   /* Caller allocation units */                                                        
			out_uint64_le(out, stat_fs.f_bfree);    /* Available allocation units */                                                     																																 
			out_uint32_le(out, stat_fs.f_bsize / 0x200);    /* Sectors per allocation unit */                                            
			out_uint32_le(out, 0x200);      /* Bytes per sector */                                                                       
        break;           
		case FileFsAttributeInformation:				   
			stmp = s_alloc(PATH_MAX * 4);
			out_utf16s_no_eos(stmp, fsinfo->type);
			s_mark_end(stmp);

			out_uint32_le(out, FS_CASE_SENSITIVE | FS_CASE_IS_PRESERVED);	/* fs attributes */
			out_uint32_le(out, F_NAMELEN(stat_fs));	/* max length of filename */

			out_uint32_le(out, s_length(stmp));	/* length of fsinfo->type string */
			out_stream(out, stmp);	/* fsinfo->typ string */

/* to be analyzed rzookol instead of 2 lines above

out_uint32_le(out, 2 * strlen(fsinfo->type));	
			rdp_out_unistr(out, fsinfo->type, 2 * strlen(fsinfo->type) - 2);

*/
			s_free(stmp);
			break;

		case FileFsLabelInformation:
		case FileFsDeviceInformation:
		case FileFsControlInformation:
		case FileFsObjectIdInformation:
		case FileFsMaximumInformation:

		default:
//KPrintF("disk_query_volume_information RD_STATUS_INVALID_PARAMETER\n");
			unimpl("IRP Query Volume Information class: 0x%x\n", info_class);
			return RD_STATUS_INVALID_PARAMETER;
	}
	
	 //KPrintF("disk_query_volume_information success \n");     
	return RD_STATUS_SUCCESS;
}


#ifdef ENABLE_AMIGA
static int fnmatch(const char *pattern, const char *string, int flags)
{
  if ('/' == *pattern)
  {
    pattern++;
  }

  if (!strcmp(string, "."))
  {
    return 1;
  }

  if (!strcmp(string, ".."))
  {
    return 1;
  }

  if (0 == pattern[0] || !strcmp(pattern, "*"))
  {
    return 0;
  }

  if (!strcmp(pattern, string))
  {
    return 0;
  }

  return 1;
}
#endif

RD_NTSTATUS disk_query_directory(RD_NTHANDLE handle, uint32 info_class, char *pattern, STREAM out)
{
	uint32 file_attributes, ft_low, ft_high;
	char *dirname, fullpath[PATH_MAX];
	BPTR pdir;
	struct FileInfoBlock *fib;
	struct fileinfo *pfinfo;
	time_t date;
	BOOL success;
	STREAM stmp;
	pfinfo = &(g_fileinfo[handle]);
	pdir = pfinfo->pdir;
	dirname = pfinfo->path;
	file_attributes = 0;

	//KPrintF("disk_query_directory patter:\"%s\"    path:%s info:%d\n ", pattern, pfinfo->path, info_class);
	
	if (!pdir) 	
	{
		out_uint8(out, 0); 
		return RD_STATUS_NO_SUCH_FILE;
	}
	
	if ( (info_class == FileBothDirectoryInformation) ||
		 (info_class == FileDirectoryInformation) ||
	     (info_class == FileFullDirectoryInformation) ||
		 (info_class == FileNamesInformation) 
	)
	{
		/* If a search pattern is received, remember this pattern, and restart search */
		if (pattern && pattern[0] != 0)
		{
			success = Examine(pdir, pfinfo->FIB);
			strncpy(pfinfo->pattern, 1 + strrchr(pattern, '/'), PATH_MAX - 1);
		}
		
		success = ExNext(pdir, pfinfo->FIB);

		while (success != DOSFALSE)
		{
			//KPrintF("A: %s\n",pfinfo->FIB->fib_FileName );
			if (!fnmatch(pfinfo->pattern, pfinfo->FIB->fib_FileName, 0))
				break;
		
			success = ExNext(pdir, pfinfo->FIB);
		}
		
		if (success == DOSFALSE)
			return RD_STATUS_NO_MORE_FILES;

		//KPrintF("Found: %s\n", pfinfo->FIB->fib_FileName);
		fib  = pfinfo->FIB;
		/* Get information for directory entry */
		date=convertDateStamp(&fib->fib_Date); 

		if ((fib->fib_DirEntryType == ST_USERDIR) || (fib->fib_DirEntryType == ST_ROOT))
				file_attributes |= FILE_ATTRIBUTE_DIRECTORY;

		if (fib->fib_FileName[0] == '.')
			file_attributes |= FILE_ATTRIBUTE_HIDDEN;

		if (!file_attributes)
			file_attributes |= FILE_ATTRIBUTE_NORMAL;
	
		/* Return requested information */
		out_uint32_le(out, 0);	/* NextEntryOffset */
		out_uint32_le(out, 0);	/* FileIndex zero */

		// Write entry name as utf16 into stmp
		stmp = s_alloc(PATH_MAX * 4);
		out_utf16s_no_eos(stmp, pfinfo->FIB->fib_FileName);
		s_mark_end(stmp);
		
		switch (info_class)
		{
			case FileBothDirectoryInformation:

			seconds_since_1970_to_filetime(date, &ft_high,
							   &ft_low);
			out_uint32_le(out, ft_low);	/* create time */
			out_uint32_le(out, ft_high);

			out_uint32_le(out, ft_low);	/* last_access_time */
			out_uint32_le(out, ft_high);

			out_uint32_le(out, ft_low);	/* last_write_time */
			out_uint32_le(out, ft_high);

			out_uint32_le(out, ft_low);	/* change_write_time */
			out_uint32_le(out, ft_high);

			out_uint32_le(out, fib->fib_Size);	/* filesize low */
			out_uint32_le(out, 0);	/* filesize high */
			out_uint32_le(out, fib->fib_Size);	/* filesize low */
			out_uint32_le(out, 0);	/* filesize high */
			out_uint32_le(out, file_attributes);
			out_uint32_le(out, s_length(stmp));	/* length of dir entry name string */
			out_uint32_le(out, 0);	/* EaSize */
			out_uint8(out, 0);	/* ShortNameLength */
			out_uint8s(out, 24);	/* ShortName (8.3 name) */
			out_stream(out, stmp); /* dir entry name string */
			
			break;
			case FileDirectoryInformation:

			seconds_since_1970_to_filetime(date, &ft_high,
						       &ft_low);
			out_uint32_le(out, ft_low);	/* create time */
			out_uint32_le(out, ft_high);

			seconds_since_1970_to_filetime(date,  &ft_high, &ft_low);
			out_uint32_le(out, ft_low);	/* last_access_time */
			out_uint32_le(out, ft_high);

			seconds_since_1970_to_filetime(date, &ft_high, &ft_low);
			out_uint32_le(out, ft_low);	/* last_write_time */
			out_uint32_le(out, ft_high);

			seconds_since_1970_to_filetime(date, &ft_high, &ft_low);
			out_uint32_le(out, ft_low);	/* change_write_time */
			out_uint32_le(out, ft_high);

		
			out_uint32_le(out, fib->fib_Size);	/* filesize low */
			out_uint32_le(out, 0);	/* filesize high */
			out_uint32_le(out, fib->fib_Size);	/* filesize low */
			out_uint32_le(out, 0);	/* filesize high */
			
			out_uint32_le(out, file_attributes);
			out_uint32_le(out, s_length(stmp));	/* dir entry name string length */
			out_stream(out, stmp); /* dir entry name */
			break;


		case FileFullDirectoryInformation:

			seconds_since_1970_to_filetime(date, &ft_high,
						       &ft_low);
			out_uint32_le(out, ft_low);	/* create time */
			out_uint32_le(out, ft_high);

			seconds_since_1970_to_filetime(date, &ft_high, &ft_low);
			out_uint32_le(out, ft_low);	/* last_access_time */
			out_uint32_le(out, ft_high);

			seconds_since_1970_to_filetime(date, &ft_high, &ft_low);
			out_uint32_le(out, ft_low);	/* last_write_time */
			out_uint32_le(out, ft_high);

			seconds_since_1970_to_filetime(date, &ft_high, &ft_low);
			out_uint32_le(out, ft_low);	/* change_write_time */
			out_uint32_le(out, ft_high);

		
			out_uint32_le(out, fib->fib_Size);	/* filesize low */
			out_uint32_le(out, 0);	/* filesize high */
			out_uint32_le(out, fib->fib_Size);	/* filesize low */
			out_uint32_le(out, 0);	/* filesize high */
			out_uint32_le(out, file_attributes);
			out_uint32_le(out, s_length(stmp));	/* dir entry name string length */
			out_uint32_le(out, 0);	/* EaSize */
			out_stream(out, stmp); /* dir entry name */
			break;


		case FileNamesInformation:

			out_uint32_le(out, s_length(stmp));	/* dir entry name string length */
			out_stream(out, stmp); /* dir entry name */
			break;
			
		default:
			//logger(Disk, Warning, "disk_query_directory(), unhandled directory info class 0x%x", info_class);
			s_free(stmp);
			return RD_STATUS_INVALID_PARAMETER;	
		}
		s_free(stmp);
	}
	else
	{
		//KPrintF("disk_query_directory(), unhandled directory info class\n");
		return RD_STATUS_INVALID_PARAMETER;
	}
	//KPrintF("succes !!\n");
	return RD_STATUS_SUCCESS;
}



static RD_NTSTATUS
disk_device_control(RD_NTHANDLE handle, uint32 request, STREAM in, STREAM out)
{
	if (((request >> 16) != 20) || ((request >> 16) != 9))
		return RD_STATUS_INVALID_PARAMETER;

	/* extract operation */
	request >>= 2;
	request &= 0xfff;

	printf("DISK IOCTL %d\n", request);

	switch (request)
	{
		case 25:	/* ? */
		case 42:	/* ? */
		default:
			unimpl("DISK IOCTL %d\n", request);
			return RD_STATUS_INVALID_PARAMETER;
	}

	return RD_STATUS_SUCCESS;
}

DEVICE_FNS disk_fns = {
	disk_create,
	disk_close,
	disk_read,
	disk_write,
	disk_device_control	/* device_control */
};
